import multiprocessing
from concurrent.futures  import ProcessPoolExecutor

def get_cpu_count():
    return multiprocessing.cpu_count()

def parallel_run(func, *iterables, max_workers):

    with ProcessPoolExecutor(max_workers) as executor:
        result = executor.map(func, *iterables)

    return [i for i in result]


import sys
import argparse

class CustomHelpFormatter(argparse.HelpFormatter):
    def _format_action(self, action):
        if type(action) == argparse._SubParsersAction:
            # inject new class variable for subcommand formatting
            subactions = action._get_subactions()
            if subactions:
                invocations = [self._format_action_invocation(a) for a in subactions]
                self._subcommand_max_length = max(len(i) for i in invocations)

        if type(action) == argparse._SubParsersAction._ChoicesPseudoAction:
            # format subcommand help line
            subcommand = self._format_action_invocation(action) # type: str
            width = self._subcommand_max_length
            help_text = ""
            if action.help:
                help_text = self._expand_help(action)
            return "  {:{width}}     {}\n".format(subcommand, help_text, width=width)

        elif type(action) == argparse._SubParsersAction:
            # process subcommand help section
            msg = '\n'
            for subaction in action._get_subactions():
                msg += self._format_action(subaction)
            return msg
        else:
            return super(CustomHelpFormatter, self)._format_action(action)

        
class CLIParser(argparse.ArgumentParser):
    
    def __init__(self, description='', usage=None, version='', **kwargs):
        
        # create default usage string
        if usage is None:
            if version:
                usage = "%(prog)s [-v|--version] [-h|--help] <command> [<args>]"
            else:
                usage = "%(prog)s [-h|--help] <command> [<args>]"
        
        custom_kwargs = {"description": description,
                         "formatter_class": CustomHelpFormatter,
                         "usage" : usage,
                         "add_help": False}
        kwargs = {**kwargs, **custom_kwargs}
        
        super().__init__(**kwargs)
        
        if version: 
            version_str = '%(prog)s version {}'.format(version)
            self.add_argument('-v', '--version',
                              action='version',
                              version=version_str,
                              help="Print program's version number")
            
        self.add_argument('-h', '--help', 
                          action="store_true",
                          help='Print help') 
        
    def set_title(self, title):
        self._positionals.title = title
        
    def error(self, message):
        '''
            Override default behavior on error to print help message instead
        '''
        sys.stderr.write('ERROR: %s\n' % message)
        self.print_help()
        sys.exit(0)
        
    def parse_known_args(self, args=None, namespace=None, **kwargs):
        # print help if no commands or arguments are given
        if (args is None and len(sys.argv) < 2):
            self.print_help()
            sys.exit(0)
        args, argv = super().parse_known_args(args, namespace, **kwargs)
        # print help if -h or --help are given
        if args.help:
            self.print_help()
            sys.exit(0)
        args.__dict__.pop('help', None)
        return args, argv
    
    def load_argument_options(self, arg_opts, **kwargs):
        for arg in arg_opts:
            arg_data = arg_opts[arg]
            full_arg = '--{}'.format(arg)
            if arg_data['abbr'] is None:
                arg_options = (full_arg,)
            else:
                abbr_arg = '-{}'.format(arg_data['abbr'])
                arg_options = (abbr_arg, full_arg)
            extra_options = {}
            extra_options['help'] = arg_data['description']
            if not isinstance(arg_data['type'], tuple):
                arg_type = tuple([arg_data['type']])
            else:
                arg_type = arg_data['type']

            if bool in arg_type:
                extra_options['action'] = 'store_true' if not arg_data['default'] else 'store_false'
            else:
                if dict in arg_type:
                    extra_options['type'] = None
                elif list in arg_type:
                    extra_options['type'] = None
                    extra_options['nargs'] = '+'
                else:
                    extra_options['type'] = arg_type[0]
                extra_options['metavar'] = ''
                extra_options['required'] = arg_data['required']
                if not arg_data['required']:
                    extra_options['default'] = arg_data['default']
            
            if 'choice' in arg_data:
                extra_options['choices'] = arg_data['choice']
                
            self.add_argument(*arg_options, **extra_options)           


import os
import time
import glob
import json
import math
from copy import deepcopy
from typing import List, Dict, Optional

import uproot
import numpy as np

from quickstats.components import CLIParser, NuisanceParameterPull
from quickstats.components.root_plots import StandardGraphMaker, set_attrib
from quickstats.components.numerics import array_swap, reorder_arrays, reverse_arrays
from quickstats.utils import root_utils

import ROOT

class NPRankingPlotter(object):
    _BEAM_ENERGY_       = 13
    _LUMINOSITY_        = 36.5
    _BOX_HEIGHT_        = 0.25
    _BOX_HEIGHT_OL_     = 0.125
    _SIGMA_TOT_LO_      = 1.0
    _SIGMA_TOT_HI_      = 1.0
    _CHANNEL_POS_X_     = 0.36
    _CHANNEL_POS_Y_     = 0.19
    _POSITION_LABEL_X_  = 0.36 
    _POSITION_LABEL_DX_ = 0.1225
    _POSITION_LABEL_Y_  = 0.185
    _POSITION_LABEL_DY_ = 0.0275
    _DEFAULT_PLOT_CONFIG_ = {
        'num_param'    : -1,
        'threshold'    : 0.0,
        'scale_poi'    : 1.0,
        'scale_theta'  : 1.0,
        'show_names'   : True,
        'show_prefit'  : True,
        'show_postfit' : True,
        'relative'     : False,
        'correlation'  : True,
        'onesided'     : True,
        'rank'         : True,
        'log_level'    : 'INFO',
        'map_param'    : '',
        'label'        : '',
        'poi'          : 'mu'
    }
    
    _DEFAULT_IMPACT_ = {
        'nuis_val'    : -999,
        'nuis_up'     : 0,
        'nuis_down'   : 0,
        'nuis_prefit' : 1.,
        'poi_hat'     : 0,
        'poi_up'      : 0,
        'poi_down'    : 0,
        'poi_nom_up'  : 0,
        'poi_nom_down': 0
    }
    
    _CLI_USAGE_       = 'quickstats plot_ranking [-h|--help] [<args>]'
    _CLI_DESCRIPTION_ = 'Tool for making NP ranking plots'   
    _CLI_ARG_OPTIONS_ = \
        {
            'input_dir': {
                'abbr'        : 'i',
                'description' : 'Directory containing the NP pulls root files',
                'required'    : True,
                'type'        : str
            },
            'poi': {
                'abbr'        : 'p',
                'description' : 'POI in the ranking',
                'required'    : False,
                'type'        : str,
                'default'     : 'mu'
            },           
            'overlay_dir': {
                'abbr'        : 'v',
                'description' : 'Directory containing the root files for overlay',
                'required'    : False,
                'type'        : str,
                'default'     : ''
            },   
            'num_param': {
                'abbr'        : 'n',
                'description' : 'Top n parameters to show',
                'required'    : False,
                'type'        : int,
                'default'     : -1
            },      
            'threshold': {
                'abbr'        : 't',
                'description' : 'Threshold on impact',
                'required'    : False,
                'type'        : float,
                'default'     : 0.0
            },      
            'scale_poi': {
                'abbr'        : None,
                'description' : 'Scaling of impact axis',
                'required'    : False,
                'type'        : float,
                'default'     : 1.0
            },          
            'scale_theta': {
                'abbr'        : None,
                'description' : 'Scaling of pull axis',
                'required'    : False,
                'type'        : float,
                'default'     : 1.0
            },   
            'show_names': {
                'abbr'        : None,
                'description' : 'Draw nuisance parameter names',
                'required'    : False,
                'type'        : int,
                'default'     : 1
            }, 
            'show_prefit': {
                'abbr'        : None,
                'description' : 'Draw prefit impact',
                'required'    : False,
                'type'        : int,
                'default'     : 1
            },  
            'show_postfit': {
                'abbr'        : None,
                'description' : 'Draw postfit impact',
                'required'    : False,
                'type'        : int,
                'default'     : 1
            },   
            'relative': {
                'abbr'        : None,
                'description' : 'Show relative variation',
                'required'    : False,
                'type'        : int,
                'default'     : 0
            },    
            'correlation': {
                'abbr'        : None,
                'description' : 'Visualize sign of correlation',
                'required'    : False,
                'type'        : int,
                'default'     : 1
            },          
            'onesided': {
                'abbr'        : None,
                'description' : 'Show one sided variations',
                'required'    : False,
                'type'        : int,
                'default'     : 1
            },        
            'rank': {
                'abbr'        : None,
                'description' : 'Order nuisances by impact',
                'required'    : False,
                'type'        : int,
                'default'     : 1
            },  
            'log_level': {
                'abbr'        : None,
                'description' : 'Log level',
                'required'    : False,
                'type'        : str,
                'default'     : "INFO"
            },  
            'map_param': {
                'abbr'        : 'm',
                'description' : 'Map to nicer parameter names',
                'required'    : False,
                'type'        : str,
                'default'     : ""
            },  
            'label': {
                'abbr'        : 'l',
                'description' : 'Label on plot',
                'required'    : False,
                'type'        : str,
                'default'     : ""
            },
            'out_dir': {
                'abbr'        : 'o',
                'description' : 'Directory where the output files are created',
                'required'    : False,
                'type'        : str,
                'default'     : "./"
            },        
        }
    
    @property
    def config(self):
        return self._config
    
    @property
    def color_scheme(self):
        return self._color_scheme

    def __init__(self):   
        self.configure()

    def get_parser(self, **kwargs):
        parser = CLIParser(description=self._CLI_DESCRIPTION_,
                           usage=self._CLI_USAGE_)
        parser.load_argument_options(self._CLI_ARG_OPTIONS_)
        return parser
    
    def run_parser(self, args=None):
        parser = self.get_parser()
        kwargs = vars(parser.parse_args(args))
        self.run_plot(**kwargs)   
        
    def configure(self, **kwargs):
        self._config = self._DEFAULT_PLOT_CONFIG_
        for k,v in kwargs.items():
            if k in self.config:
                self._config[k] = v
            else:
                raise ValueError('unrecognized plot setting: {}'.format(k))                

    def get_plot_options(self):
        anticorr = self.config['correlation']
        kColorStandardBand   = "#ff0000"
        kColorStandardBandOL = "#950098"
        kColorPulls          = "#9d9d9d"
        kColorPrefit         = "#000155"
        kColorPrefitOL       = "#027734"
        kColorPostfit        = "#0094ff"
        kColorPostfitOL      = "#44ff94"
        kColorAntiCorr       = kColorPostfitOL if anticorr else kColorPostfit
        kColorAntiCorrOL     = kColorPostfit if anticorr else kColorPostfitOL
        kColorAntiCorrNom    = kColorPrefitOL if anticorr else kColorPrefit
        kColorAntiCorrNomOL  = kColorPrefit if anticorr else kColorPrefitOL
        plot_options = {
            'nuis_pulls':{
                'LineColor'        : ROOT.kBlack,
                'MarkerColor'      : ROOT.kBlack,
                'MarkerStyle'      : 20,
                'LineStyle'        : 1,
                'LineWidth'        : 1,
                'MarkerSize'       : 1,
                'Xaxis.TitleOffset': 1.2
            },
            'nuis_pulls_ol':{
                'LineColor'        : ROOT.kBlack,
                'MarkerColor'      : ROOT.kBlack,
                'MarkerStyle'      : 24,
                'LineStyle'        : 1,
                'LineWidth'        : 2,
                'MarkerSize'       : 1,
                'Xaxis.TitleOffset': 1.2
            },
            'one_sigma_box':{
                'LineColor'        : ROOT.TColor.GetColor(kColorStandardBand),
                'MarkerColor'      : ROOT.TColor.GetColor(kColorStandardBand),
                'LineStyle'        : 1,
                'LineWidth'        : 1,
                'MarkerSize'       : 1.25,
                'Xaxis.TitleOffset': 1.2
            },  
            'one_sigma_box_ol':{
                'LineColor'        : ROOT.TColor.GetColor(kColorStandardBandOL),
                'MarkerColor'      : ROOT.TColor.GetColor(kColorStandardBandOL),
                'LineStyle'        : 1,
                'LineWidth'        : 3,
                'MarkerSize'       : 1.25,
                'Xaxis.TitleOffset': 1.2
            },
            'line':{
                'LineColor'        : ROOT.TColor.GetColor(kColorPulls),  
                'LineStyle'        : 2,
                'LineWidth'        : 2
            },
            'poi_postfit_corr':{
                'LineColor'        : ROOT.TColor.GetColor(kColorPostfit),
                'FillColor'        : ROOT.TColor.GetColor(kColorPostfit),
                'LineWidth'        : 0,
                'MarkerSize'       : 0,
            },
            'poi_postfit_ol_corr':{
                'LineColor'        : ROOT.TColor.GetColor(kColorPostfitOL),
                'FillColor'        : ROOT.TColor.GetColor(kColorPostfitOL),
                'LineWidth'        : 0,
                'MarkerSize'       : 0,
            },
            'poi_postfit_anticorr':{
                'LineColor'        : ROOT.TColor.GetColor(kColorAntiCorr),
                'FillColor'        : ROOT.TColor.GetColor(kColorAntiCorr),
                'LineWidth'        : 0,
                'MarkerSize'       : 0,
            },
            'poi_postfit_anticorr_ol':{
                'LineColor'        : ROOT.TColor.GetColor(kColorAntiCorrOL),
                'FillColor'        : ROOT.TColor.GetColor(kColorAntiCorrOL),
                'LineWidth'        : 0,
                'MarkerSize'       : 0,
            },            
            'poi_prefit_corr':{
                'LineColor'        : ROOT.TColor.GetColor(kColorPrefit),
                'FillColor'        : ROOT.TColor.GetColor(kColorPrefit),
                'FillStyle'        : 0,
                'LineWidth'        : 1,
                'MarkerSize'       : 0,
            },
            'poi_prefit_ol_corr':{
                'LineColor'        : ROOT.TColor.GetColor(kColorPrefitOL),
                'FillColor'        : ROOT.TColor.GetColor(kColorPrefitOL),
                'FillStyle'        : 0,
                'LineWidth'        : 1,
                'MarkerSize'       : 0,
            },
            'poi_prefit_anticorr':{
                'LineColor'        : ROOT.TColor.GetColor(kColorAntiCorrNom),
                'FillColor'        : ROOT.TColor.GetColor(kColorAntiCorrNom),
                'FillStyle'        : 0,
                'LineWidth'        : 1,
                'MarkerSize'       : 0,
            },
            'poi_prefit_ol_anticorr':{
                'LineColor'        : ROOT.TColor.GetColor(kColorAntiCorrNomOL),
                'FillColor'        : ROOT.TColor.GetColor(kColorAntiCorrNomOL),
                'FillStyle'        : 0,
                'LineWidth'        : 1,
                'MarkerSize'       : 0,
            },
            'shade':{
                'LineColor'        : 18,
                'FillColor'        : 18,
                'FillStyle'        : 3001,
                'LineWidth'        : 1,
                'MarkerSize'       : 0
            },
            'label':{
                'Xaxis.LabelColor' : ROOT.kWhite,
                'Xaxis.AxisColor'  : ROOT.kWhite,
                'Yaxis.LabelColor' : ROOT.kBlack,
                'Yaxis.AxisColor'  : ROOT.kBlack,
                'Yaxis.TickLength' : 0.,
                'Stats'            : 0
            },
            'axis_poi':{
                'Name'             : 'axis_poi',
                'Title'            : "#Delta#hat{#mu}/#Delta#hat{#mu}_{tot}" if self.config['relative'] else \
                                     "#Delta#hat{" + self.config['poi'] + "}",
                'TitleOffset'      : 1.1,
                'LineColor'        : ROOT.kBlack,
                'LabelColor'       : ROOT.kBlack,
                'TitleColor'       : ROOT.kBlack,
                'LabelSize'        : 0.034,
                'TitleSize'        : 0.034
            },
            'axis_theta':{
                'Name'             : 'axis_theta',
                'Title'            : "(#hat{#theta} - #theta_{0})/#Delta#theta",
                'TitleOffset'      : 1.1,
                'LineColor'        : ROOT.kBlack,
                'LabelColor'       : ROOT.kBlack,
                'TitleColor'       : ROOT.kBlack,
                'LabelSize'        : 0.034,
                'TitleSize'        : 0.034
            },        
            'axis_label':{
                'LineColor'        : ROOT.kBlack,
                'TitleColor'       : ROOT.kWhite,
                'LabelSize'        : 0,
                'Ndivisions'       : 0
            },
            'legend':{
                'FillStyle'        : 0,
                'TextSize'         : 0.0225,
                'BorderSize'       : 0,
                'FillColor'        : 0
            },
            'ATLAS_label':{
                'NDC'              : None,
                'TextFont'         : 72,
                'TextColor'        : ROOT.kBlack,
                'TextSize'         : 0.035
            },
            'internal_label':{
                'NDC'              : None,
                'TextFont'         : 42,
                'TextColor'        : 1,
                'TextSize'         : 0.035
            }            
        }
        return plot_options                
        
    def setup_canvas(self):
        
        if not ROOT.gROOT.FindObject("c1"):
            canvas = ROOT.TCanvas("c1", "c1", 1500, 2000)
        else:
            canvas = ROOT.gROOT.FindObject("c1")
        if not ROOT.gROOT.FindObject("pad1"): 
            pad    = ROOT.TPad("pad1", "pad1", 0.0, 0.0, 1.0, 1.0, 0)
        else:
            pad    = ROOT.gROOT.FindObject("pad1")
            
        if self.config['show_names']:
            pad.SetLeftMargin(0.35)
        else:
            pad.SetLeftMargin(0.05)
        pad.SetRightMargin(0.05)
        pad.SetBottomMargin(0.09)
        pad.SetTopMargin(0.09)
        
        pad.Draw()
        pad.cd()
        return canvas, pad
        
    @staticmethod
    def extract_pull_results(input_dir:str, poi:str):
        if not os.path.exists(input_dir):
            raise FileNotFoundError('input directory {} does not exist'.format(input_dir))
        if not os.path.isdir(input_dir):
            raise ValueError('{} is not a directory'.format(input_dir))
        root_files = glob.glob(os.path.join(input_dir, "*.root"))
        pull_results = {
            'nuis_label'  : [],
            'nuis_val'    : [],
            'nuis_up'     : [],
            'nuis_down'   : [],
            'nuis_prefit' : [],
            'poi_hat'     : [],
            'poi_up'      : [],
            'poi_down'    : [],
            'poi_nom_up'  : [],
            'poi_nom_down': []
        }
        for file in root_files:
            raw_result = NuisanceParameterPull.parse_root_result(file)
            content = raw_result.get('result', None)
            if not content:
                raise ValueError('{} is not a valid root file containing NP pulls result'.format(file))
            pull_results['nuis_label'].append(content['nuisance'])
            pull_results['nuis_val'].append(content['nuis_hat'] - content['nuis_nom'])
            pull_results['nuis_up'].append(content['nuis_hi'])
            pull_results['nuis_down'].append(content['nuis_lo'])
            pull_results['nuis_prefit'].append(content['nuis_prefit'])
            pull_results['poi_hat'].append(content['{}_hat'.format(poi)])
            pull_results['poi_up'].append(content['{}_up'.format(poi)])
            pull_results['poi_down'].append(content['{}_down'.format(poi)])
            pull_results['poi_nom_up'].append(content['{}_up_nom'.format(poi)])
            pull_results['poi_nom_down'].append(content['{}_down_nom'.format(poi)])
        # convert to numpy arrays for easier post-processing
        for key in pull_results:
            pull_results[key] = np.array(pull_results[key])
        return pull_results
            
    def run_plot(self, input_dir:str, poi:str, overlay_dir:str='', num_param:int=-1, threshold:float=0.0,
                 scale_poi:float=1.0, scale_theta:float=1.0, show_names:bool=True, show_prefit:bool=True,
                 show_postfit:bool=True, relative:bool=False, correlation:bool=True, onesided:bool=True,
                 rank:bool=True, log_level:str="INFO", map_param:str="", label:str="", out_dir:str="./"):
        
        data = self.extract_pull_results(input_dir, poi)
        data_overlay = None
        if overlay_dir:
            data_overlay = self.extract_pull_results(overlay_dir, poi)
        
        param_dict = None
        if map_param:
            if os.path.exists(map_param):
                with open(map_param) as f:
                    param_dict = json.load(f)
            else:
                raise FileNotFoundError("map file {} does not exist".format(map_param))
        
        self.configure(num_param=num_param, threshold=threshold, scale_poi=scale_poi, rank=rank,
                       scale_theta=scale_theta, show_names=show_names, relative=relative,
                       correlation=correlation, onesided=onesided, label=label,
                       show_prefit=show_prefit, show_postfit=show_postfit, 
                       log_level=log_level, map_param=map_param, poi=poi)
        
        if data_overlay:
            all_nuis_labels = set(data['nuis_label']).union(set(data_overlay['nuis_label']))
        else:
            all_nuis_labels = None
        
        data = self.process_data(data, **self.config, pad_labels=all_nuis_labels)
        labels_to_keep = data['nuis_label']
        
        if data_overlay:
            overlay_config = deepcopy(self.config)
            overlay_config['threshold'] = 0.0
            overlay_config['num_param'] = -1
            data_overlay = self.process_data(data_overlay, **overlay_config,
                                             pad_labels=all_nuis_labels)
            # remove extra labels not present in data
            keep_indices = np.in1d(data_overlay['nuis_label'], labels_to_keep)
            for key in data_overlay:
                data_overlay[key] = data_overlay[key][keep_indices]
        #return data, data_overlay
        self.create_ranking_plot(data, data_overlay, param_dict, out_dir=out_dir)
    
    @staticmethod
    def pad_default_impact(source, pad_labels:Optional[List]=None, 
                           default_impact:Optional[Dict]=None):
        # no extra impacts for padding, just exit
        if not pad_labels:
            return None
        if default_impact is None:
            default_impact = NPRankingPlotter._DEFAULT_IMPACT_        
        extra_labels = np.array(list(set(pad_labels) - set(source['nuis_label'])))
        n_extra = len(extra_labels)
        if n_extra == 0:
            return None
        new_impact = {'nuis_label': extra_labels}
        new_impact.update({key: np.array([default_impact[key]]*n_extra) for key in default_impact})
        for key in source:
            if key not in new_impact:
                raise ValueError('missing default value for the attribute {}'.format(key))
            source[key] = np.concatenate((source[key], new_impact[key]))
            
    @staticmethod
    def remove_impact_data(data, threshold:float=0.0, num_param:int=-1):
        # remove impacts below threshold
        sigma_tol = NPRankingPlotter._SIGMA_TOT_LO_ + NPRankingPlotter._SIGMA_TOT_HI_
        below_threshold = np.where(((data['poi_up'] + data['poi_down']) / sigma_tol) < threshold)
        for i in below_threshold[0]:
            print('WARNING: Removing "{}" (impact up:{}, impact down:{}) below threshold {}.'.format(
                  data['nuis_label'][i], data['poi_up'][i], data['poi_down'][i], threshold))
        for key in data:
            data[key] = np.delete(data[key], below_threshold)
        # retain only top num_param impacts
        if num_param > 0:
            for key in data:
                data[key] = data[key][:num_param]
    
    @staticmethod
    def process_impact(impact_up:np.ndarray, impact_down:np.ndarray, 
                       correlation:bool=True, onesided:bool=True):
        if correlation:
            array_swap(impact_up, impact_down, (impact_up < 0) & (impact_down > 0))

        if (not correlation) or (not onesided):
            array_swap(impact_up, impact_down, impact_up < 0)
        # secondary rows for one-sided impacts
        impact_up_sec = deepcopy(impact_up)
        impact_down_sec = deepcopy(impact_down)
        if onesided:
            indices = (impact_up > 0) & (impact_down > 0)
            impact_up[indices] = impact_down_sec[indices]
            impact_down[indices] = 0
            impact_down_sec[indices] = 0
            indices = (impact_up < 0) & (impact_down < 0)
            impact_down_sec[indices] = impact_up[indices]
            impact_up[indices] = 0
            impact_up_sec[indices] = 0
        impact_up[:], impact_down[:] = abs(impact_up), abs(impact_down)
        impact_up_sec[:], impact_down_sec[:] = abs(impact_up_sec), abs(impact_down_sec)
        return impact_up_sec, impact_down_sec
    
    @staticmethod
    def process_data(data, scale_theta:float=1.0, scale_poi:float=1.0, correlation:bool=True,
                     onesided:bool=True, ranking:bool=True, relative:bool=False, threshold:float=0.0,
                     num_param:int=-1, pad_labels:Optional[List]=None, **kwargs):
        source = deepcopy(data)
        # subtract poi_hat offset from impacts
        for key in ['poi_up', 'poi_down', 'poi_nom_up', 'poi_nom_down']:
            source[key] -= source['poi_hat']   
        anticorrelated     = source['nuis_label'][(source['poi_up'] < 0) & (source['poi_down'] > 0)]
        anticorrelated_nom = source['nuis_label'][(source['poi_nom_up'] < 0) & (source['poi_nom_down'] > 0)]
        poi_up_sec, poi_down_sec = NPRankingPlotter.process_impact(
            source['poi_up'], source['poi_down'], correlation, onesided)
        source['poi_up_sec'] = poi_up_sec
        source['poi_down_sec'] = poi_down_sec
        poi_nom_up_sec, poi_nom_down_sec = NPRankingPlotter.process_impact(
            source['poi_nom_up'], source['poi_nom_down'], correlation, onesided)
        source['poi_nom_up_sec'] = poi_nom_up_sec
        source['poi_nom_down_sec'] = poi_nom_down_sec        
        max_poi = max(max(source['poi_up']), max(source['poi_down']),
                      max(source['poi_up_sec']), max(source['poi_down_sec']))
        NPRankingPlotter.pad_default_impact(source, pad_labels)
        if ranking:
            impact = (source['poi_up'] + source['poi_down'] +
                     source['poi_up_sec'] + source['poi_down_sec'])
            impact /= (1*(source['poi_up'] != 0) + 1*(source['poi_down'] != 0) +
                       1*(source['poi_up_sec'] != 0) + 1*(source['poi_down_sec'] != 0))
            # order by impact, from largest to smallest
            reorder_arrays(impact, *source.values())
        else:
            # order by NP names
            reorder_arrays(source['nuis_label'], *source.values(), descending=False)

        NPRankingPlotter.remove_impact_data(source, threshold, num_param)
        
        # rescaling impacts and pulls
        for key in ['poi_up', 'poi_down', 'poi_nom_up', 'poi_nom_down',
                    'poi_up_sec', 'poi_down_sec', 'poi_nom_up_sec', 'poi_nom_down_sec']:
            source[key] *= (scale_poi / max_poi)
            if relative:
                source[key] /= source['poi_hat']
        source['nuis_val'] *= scale_theta
        for key in ['nuis_up', 'nuis_down']:
            source[key] /= source['nuis_prefit']   
            source[key] = abs(source[key])*scale_theta
    
        source['poi_corr']     = ~np.in1d(source['nuis_label'], anticorrelated)
        source['poi_nom_corr'] = ~np.in1d(source['nuis_label'], anticorrelated_nom)
        # dirty fix for carrying max_poi inside source
        source['max_poi'] = np.array([max_poi]*len(source['nuis_label']))
        if ranking:
            # order by impact, from smallest to largest
            reverse_arrays(*source.values())
        return source
    
    @staticmethod
    def split_impact_by_correlation(impact_up, impact_down, correlated_indices):
        impacts = {
            'up_corr'       : impact_up.copy(),
            'up_anticorr'   : impact_up.copy(),
            'down_corr'     : impact_down.copy(),
            'down_anticorr' : impact_down.copy(),
        }
        impacts['up_corr'][~correlated_indices]      = 0.0
        impacts['up_anticorr'][correlated_indices]   = 0.0
        impacts['down_corr'][~correlated_indices]    = 0.0
        impacts['down_anticorr'][correlated_indices] = 0.0
        #for key in impacts:
        #    impacts[key] = np.vstack((impacts[key], impacts[key])).ravel('F')
        return impacts
    
    @staticmethod
    def _combine_impacts(primary, secondary):
        impacts = {}
        for key in ['up_corr', 'up_anticorr', 'down_corr', 'down_anticorr']:
            impacts[key] = np.vstack((primary[key], secondary[key])).ravel('F')
        return impacts
    
    def get_plot_data(self, data, data_overlay=None):
        
        n_nuis = len(data['nuis_label'])
        if data_overlay:
            n_nuis_ol = len(data['data_overlay'])
            assert n_nuis == n_nuis_ol
            
        box_up      = np.array([self.config['scale_theta']]*n_nuis)
        box_down    = np.array([self.config['scale_theta']]*n_nuis)
        box_up_ol   = data['nuis_prefit']*self.config['scale_theta']
        box_down_ol = data['nuis_prefit']*self.config['scale_theta']
        cen_up      = np.array([self._BOX_HEIGHT_]*2*n_nuis)
        cen_down    = np.array([self._BOX_HEIGHT_]*2*n_nuis)
        cen_up_ol   = np.array([self._BOX_HEIGHT_OL_]*2*n_nuis)
        cen_down_ol = np.array([self._BOX_HEIGHT_OL_]*2*n_nuis)        
             
        primary     = self.split_impact_by_correlation(data['poi_up'], 
                                                       data['poi_down'],
                                                       data['poi_corr'])
        
        secondary   = self.split_impact_by_correlation(data['poi_up_sec'], 
                                                       data['poi_down_sec'],
                                                       data['poi_corr'])        
        poi_impacts = self._combine_impacts(primary, secondary)
        
        nom_primary = self.split_impact_by_correlation(data['poi_nom_up'],
                                                       data['poi_nom_down'],
                                                       data['poi_nom_corr'])
        
        nom_secondary = self.split_impact_by_correlation(data['poi_nom_up_sec'], 
                                                         data['poi_nom_down_sec'],
                                                         data['poi_nom_corr'])   
        poi_nom_impacts = self._combine_impacts(nom_primary, nom_secondary)
        #from pdb import set_trace
        #set_trace()
        if data_overlay:
            nuis_y = np.array(range(n_nuis)) + 0.75
            poi_y  = np.array([0.625, 0.875] * n_nuis) + np.vstack((range(n_nuis), range(n_nuis))).ravel('F')
        else:
            nuis_y = np.array(range(n_nuis)) + 0.5
            poi_y  = np.array([0.25, 0.75] * n_nuis) + np.vstack((range(n_nuis), range(n_nuis))).ravel('F')

        #  make plot of pulls for nuisance parameters
        plot_data = {}
        plot_data['nuis_pulls'] = {
            'n'      : n_nuis,
            'x'      : data['nuis_val'],
            'y'      : nuis_y,
            'xerr_lo': data['nuis_down'],
            'xerr_hi': data['nuis_up']
        }
        plot_data['one_sigma_box'] = {
            'n'      : n_nuis,
            'x'      : data['nuis_val'],
            'y'      : nuis_y,
            'xerr_lo': box_down,
            'xerr_hi': box_up
        }
        plot_data['poi_prefit_corr'] = {
            'n'      : 2*n_nuis,
            'x'      : np.zeros(2*n_nuis),
            'y'      : poi_y,
            'xerr_lo': poi_nom_impacts['down_corr'],
            'xerr_hi': poi_nom_impacts['up_corr'],
            'yerr_lo': cen_down,
            'yerr_hi': cen_up,
        }
        plot_data['poi_prefit_anticorr'] = {
            'n'      : 2*n_nuis,
            'x'      : np.zeros(2*n_nuis),
            'y'      : poi_y,
            'xerr_lo': poi_nom_impacts['down_anticorr'],
            'xerr_hi': poi_nom_impacts['up_anticorr'],
            'yerr_lo': cen_down,
            'yerr_hi': cen_up,        
        }    
        plot_data['poi_postfit_corr'] = {
            'n'      : 2*n_nuis,
            'x'      : np.zeros(2*n_nuis),
            'y'      : poi_y,
            'xerr_lo': poi_impacts['down_corr'],
            'xerr_hi': poi_impacts['up_corr'],
            'yerr_lo': cen_down,
            'yerr_hi': cen_up,             
        }
        plot_data['poi_postfit_anticorr'] = {
            'n'      : 2*n_nuis,
            'x'      : np.zeros(2*n_nuis),
            'y'      : poi_y,
            'xerr_lo': poi_impacts['down_anticorr'],
            'xerr_hi': poi_impacts['up_anticorr'],
            'yerr_lo': cen_down,
            'yerr_hi': cen_up,          
        }

        if data_overlay:
            poi_overlay_impacts     = self.split_impact_by_correlation(data_overlay['poi_up'], 
                                                                       data_overlay['poi_down'], 
                                                                       data_overlay['poi_corr'])
            poi_overlay_nom_impacts = self.split_impact_by_correlation(data_overlay['poi_nom_up'],
                                                                       data_overlay['poi_nom_down'],
                                                                       data_overlay['poi_nomcorr']) 
            nuis_y        = np.array(range(n_nuis)) + 0.25
            poi_y_overlay = np.array([0.125, 0.375]*n_nuis) + np.vstack((range(n_nuis), range(n_nuis))).ravel('F')
            plot_data['nuis_pulls_ol'] = {
                'n'      : n_nuis,
                'x'      : data_overlay['nuis_val'],
                'y'      : nuis_y,
                'xerr_lo': data_overlay['nuis_down'],
                'xerr_hi': data_overlay['nuis_up']
            }
            plot_data['one_sigma_box_ol'] = {
                'n'      : n_nuis,
                'x'      : source['nuis_val'],
                'y'      : nuis_y,
                'xerr_lo': box_down_ol,
                'xerr_hi': box_up_ol
            }
            plot_data['poi_prefit_ol_corr'] = {
                'n'      : 2*n_nuis,
                'x'      : np.zeros(2*n_nuis),
                'y'      : poi_y_overlay,
                'xerr_lo': poi_overlay_nom_impacts['down_corr'],
                'xerr_hi': poi_overlay_nom_impacts['up_corr'],
                'yerr_lo': cen_down_ol,
                'yerr_hi': cen_up_ol,
            }
            plot_data['poi_prefit_ol_anticorr'] = {
                'n'      : 2*n_nuis,
                'x'      : np.zeros(2*n_nuis),
                'y'      : poi_y_overlay,
                'xerr_lo': poi_overlay_nom_impacts['down_anticorr'],
                'xerr_hi': poi_overlay_nom_impacts['up_anticorr'],
                'yerr_lo': cen_down_ol,
                'yerr_hi': cen_up_ol,        
            }    
            plot_data['poi_postfit_ol_corr'] = {
                'n'      : 2*n_nuis,
                'x'      : np.zeros(2*n_nuis),
                'y'      : poi_y_overlay,
                'xerr_lo': poi_overlay_impacts['down_corr'],
                'xerr_hi': poi_overlay_impacts['up_corr'],
                'yerr_lo': cen_down_ol,
                'yerr_hi': cen_up_ol,             
            }
            plot_data['poi_postfit_ol_anticorr'] = {
                'n'      : 2*n_nuis,
                'x'      : np.zeros(2*n_nuis),
                'y'      : poi_y_overlay,
                'xerr_lo': poi_overlay_impacts['down_anticorr'],
                'xerr_hi': poi_overlay_impacts['up_anticorr'],
                'yerr_lo': cen_down_ol,
                'yerr_hi': cen_up_ol,          
            }           
        else:
            plot_data['nuis_pulls_ol']           = None
            plot_data['one_sigma_box_ol']        = None
            plot_data['poi_prefit_ol_corr']      = None
            plot_data['poi_prefit_ol_anticorr']  = None
            plot_data['poi_postfit_ol_corr']     = None
            plot_data['poi_postfit_ol_anticorr'] = None
        n_shade = (n_nuis + 1)//2
        plot_data['shade'] = {
            'n'      : n_shade,
            'x'      : np.zeros(n_shade),
            'y'      : np.array(range(n_shade))*2 + 0.5,
            'xerr_lo': np.array([100.]*n_shade),
            'xerr_hi': np.array([100.]*n_shade),
            'yerr_lo': np.array([0.5]*n_shade),
            'yerr_hi': np.array([0.5]*n_shade)
        }
        max_poi   = data['max_poi'][0]
        offset    = math.ceil(2 * n_nuis / 10)
        border_lo = -self._SIGMA_TOT_LO_ / max_poi
        border_hi = self._SIGMA_TOT_HI_ / max_poi
        plot_data['label'] = {
            "n_x"  : 1,
            "x_lo" : border_lo,
            "x_hi" : border_hi,
            "n_y"  : n_nuis + offset + 1,
            "y_lo" : -offset,
            "y_hi" : n_nuis + 1
        }      
        plot_data['axis_poi'] = {
            "x_min" : border_lo,
            "y_min" : n_nuis + 1,
            "x_max" : border_hi,
            "y_max" : n_nuis + 1,
            "w_min" : -self._SIGMA_TOT_LO_ / self.config['scale_poi'],
            "w_max" : self._SIGMA_TOT_HI_ / self.config['scale_poi'],
            "n_div" : 510,
            "chopt" : "-L"
        }    
        plot_data['axis_theta'] = {
            "x_min" : border_lo,
            "y_min" : -offset,
            "x_max" : border_hi,
            "y_max" : -offset,
            "w_min" : (-self._SIGMA_TOT_LO_ / max_poi) / self.config['scale_theta'],
            "w_max" : (self._SIGMA_TOT_HI_ / max_poi) / self.config['scale_theta'],
            "n_div" : 510,
            "chopt" : "+R"
        }  
        plot_data['axis_label'] = {
            "x_min" : border_lo,
            "y_min" : 0,
            "x_max" : border_lo,
            "y_max" : n_nuis + 1,
            "w_min" : 0,
            "w_max" : n_nuis + 1,
            "n_div" : 0,
            "chopt" : "-R"
        }     
        plot_data['legend'] = {
            'x1'     : self._CHANNEL_POS_X_ + 0.28,
            'y1'     : self._CHANNEL_POS_Y_ - 0.0775,
            'x2'     : self._CHANNEL_POS_X_ + 0.68,
            'y2'     : self._CHANNEL_POS_Y_ + 0.02,
            "header" : "",
            "option" : "NDC"
        }
        return plot_data

    def create_ranking_plot(self, data, data_overlay=None, param_dict=None, out_dir="./"):
        
        canvas, pad = self.setup_canvas()
        n_nuis = len(data['nuis_label'])
        plot_data = self.get_plot_data(data, data_overlay)  
        plot_options = self.get_plot_options()
        
        # format histogram for labels
        hist = StandardGraphMaker.TH2F("h_ranking", "", **plot_data['label'])
        labels = data['nuis_label'] if self.config['show_names'] else np.array([""]*n_nuis)
        if param_dict is not None:
            labels = np.array([param_dict[label] for label in labels])
        offset = math.ceil(2 * n_nuis / 10)
        for i, label in enumerate(labels):
            hist.GetYaxis().SetBinLabel(i + offset + 1, label)
        label_size = (1.0 / n_nuis) if (1.0 / n_nuis) < 0.02 else 0.02
        hist.SetLabelSize(label_size, "Y")
        set_attrib(hist, **plot_options['label'])
        hist.Draw("h")
        
        # format axis for poi
        axis_poi = StandardGraphMaker.TGaxis(**plot_data['axis_poi'])
        axis_poi.ImportAxisAttributes(hist.GetXaxis())
        set_attrib(axis_poi, **plot_options['axis_poi'])
        
        # format axis for theta
        axis_theta = StandardGraphMaker.TGaxis(**plot_data['axis_theta'])
        axis_theta.ImportAxisAttributes(hist.GetXaxis())
        set_attrib(axis_theta, **plot_options['axis_theta'])
        
        # format axis for labels
        axis_label = StandardGraphMaker.TGaxis(**plot_data['axis_label'], options=plot_options['axis_label'])
        
        # format line
        line = StandardGraphMaker.TLine(options=plot_options['line'])  
        
        graphs_to_draw = ['shade']
        if self.config['show_postfit']:
            graphs_to_draw.append('poi_postfit_corr')
            graphs_to_draw.append('poi_postfit_anticorr')
            if data_overlay:
                graphs_to_draw.append('poi_postfit_ol_corr')
                graphs_to_draw.append('poi_postfit_ol_anticorr')
        if self.config['show_prefit']:
            graphs_to_draw.append('poi_prefit_corr')
            graphs_to_draw.append('poi_prefit_anticorr')
            if data_overlay:
                graphs_to_draw.append('poi_prefit_ol_corr')
                graphs_to_draw.append('poi_prefit_ol_anticorr')
        # draw graphs
        graphs = {}
        for name in graphs_to_draw:
            graphs[name] = StandardGraphMaker.GraphAsymErrors(**plot_data[name], options=plot_options[name])
            graphs[name].Draw("p2")
            
        # draw axis
        if self.config['show_postfit'] or self.config['show_prefit']:
            axis_poi.Draw()
        axis_theta.Draw()
        axis_label.Draw()
        
        # draw +-1 and 0 sigma lines for pulls
        line.DrawLine(0., 0., 0., n_nuis);
        line.DrawLine(1. * self.config['scale_theta'], 0., 1. * self.config['scale_theta'], n_nuis);
        line.DrawLine(-1. * self.config['scale_theta'], 0., -1. * self.config['scale_theta'], n_nuis);

        ROOT.gStyle.SetEndErrorSize(5.0)
        if data_overlay:
            graphs_to_draw = ['one_sigma_box', 'one_sigma_box_ol', 'nuis_pulls', 'nuis_pulls_ol']
        else:
            graphs_to_draw = ['one_sigma_box', 'nuis_pulls']
            
        for name in graphs_to_draw:
            graphs[name] = StandardGraphMaker.GraphAsymErrors(**plot_data[name], options=plot_options[name])
            graphs[name].Draw("p")
        
        canvas.SetLogy(0)
        
        # draw legends
        legend = StandardGraphMaker.TLegend(**plot_data['legend'], options=plot_options['legend'])
        legend.AddEntry(graphs['nuis_pulls'], "Pull", "lp")
        if data_overlay:
            legend.AddEntry(graphs['nuis_pulls_ol'], "Alt pull", "lp")
        legend.AddEntry(graphs['one_sigma_box'], "1 standard deviation", "l")
        if data_overlay:
            legend.AddEntry(graphs['one_sigma_box_ol'], "Alt 1 standard deviation", "l")    
        if self.config['show_prefit']:
            legend.AddEntry(graphs['poi_prefit_corr'], 
                            "Prefit Impact on #hat{" + self.config['poi'] + "}", "f")
            if data_overlay:
                legend.AddEntry(graphs['poi_prefit_corr_ol'], 
                                "Alt Prefit Impact on #hat{" + self.config['poi'] + "}", "f")
        if self.config['show_postfit']:
            legend.AddEntry(graphs['poi_postfit_corr'], 
                            "Postfit Impact on #hat{" + self.config['poi'] + "}", "f")
            if data_overlay:
                legend.AddEntry(graphs['poi_postfit_corr_ol'], 
                                "Alt Postfit Impact on #hat{" + self.config['poi'] + "}", "f")
        legend.Draw()
        
        label_1 = StandardGraphMaker.TLatex(options=plot_options['ATLAS_label'])
        label_2 = StandardGraphMaker.TLatex(options=plot_options['internal_label'])
        
        label_1.DrawLatex(self._POSITION_LABEL_X_, self._POSITION_LABEL_Y_, "ATLAS")
        label_2.DrawLatex(self._POSITION_LABEL_X_ + self._POSITION_LABEL_DX_,
                              self._POSITION_LABEL_Y_, "Internal")
        label_2.DrawLatex(self._POSITION_LABEL_X_,
                              self._POSITION_LABEL_Y_ - 1 * self._POSITION_LABEL_DY_,
                              self.config['label'])  
        label_2.DrawLatex(self._POSITION_LABEL_X_,
                              self._POSITION_LABEL_Y_ - 2 * self._POSITION_LABEL_DY_,
                              "Rank 1 to {}".format(n_nuis))         
        
        save_name = "ranking_{}_rank_{:04d}_to_{:04d}".format(self.config['poi'], 1, n_nuis)
        save_path = os.path.join(out_dir, save_name)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        for extension in ['eps', 'pdf', 'png', 'C']:
            canvas.SaveAs("{}.{}".format(save_path, extension))


##################################################################################################
# Based on https://gitlab.cern.ch/atlas-physics/stat/tools/StatisticsTools by Stefan Gadatsch
# Author: Alkaid Cheng
# Email: chi.lung.cheng@cern.ch
##################################################################################################
import os
import math
from typing import List, Optional

import ROOT

from quickstats.components.numerics import is_integer

class ExtendedModel(object):
    def __init__(self, model_name:str, fname:str, ws_name:str,
                 model_config_name:str, data_name:str, snapshot_names:List[str],
                 binned_likelihood:bool=True, tag_as_measurement:str="pdf_",
                 fix_cache:bool=True, fix_multi:bool=True,
                 interpolation_code:int=-1):
        self.fname = fname
        self.name = model_name
        self.ws_name = ws_name
        self.model_config_name = model_config_name
        self.data_name = data_name
        self.snapshot_names = snapshot_names
        self.binned_likelihood = binned_likelihood
        self.tag_as_measurement = tag_as_measurement
        self.fix_cache = fix_cache
        self.fix_multi = fix_multi
        self.interpolation_code = interpolation_code
        
        self.initialize()
  
    @property
    def file(self):
        return self._file
    @property
    def workspace(self):
        return self._workspace
    @property
    def model_config(self):
        return self._model_config
    @property
    def pdf(self):
        return self._pdf
    @property
    def data(self):
        return self._data
    @property
    def nuisance_parameters(self):
        return self._nuisance_parameters
    @property
    def global_observables(self):
        return self._global_observables
    @property
    def pois(self):
        return self._pois
    @property
    def observables(self):
        return self._observables        
        
    @staticmethod
    def modify_interp_codes(ws, interp_code, classes=None):
        if classes is None:
            classes = [ROOT.RooStats.HistFactory.FlexibleInterpVar, ROOT.PiecewiseInterpolation]
        for component in ws.components():
            for cls in classes:
                if (component.IsA() == cls.Class()):
                    component.setAllInterpCodes(interp_code)
                    class_name = cls.Class_Name().split('::')[-1]
                    print('INFO: {} {} interpolation code set to {}'.format(component.GetName(),
                                                                            class_name,
                                                                            interp_code))
        return None

    @staticmethod
    def activate_binned_likelihood(ws):
        print('INFO: Activating binned likelihood evaluation')
        for component in ws.components():
            if (component.IsA() == ROOT.RooRealSumPdf.Class()):
                component.setAttribute('BinnedLikelihood')
                print('INFO: Activated binned likelihood attribute for {}'.format(component.GetName()))
        return None
                          
    @staticmethod
    def set_measurement(ws, condition):
        print('INFO: Activating measurements to reduce memory consumption')
        for component in ws.components():
            name = component.GetName()
            if ((component.IsA() == ROOT.RooAddPdf.Class()) and condition(name)):
                component.setAttribute('MAIN_MEASUREMENT')
                print('INFO: Activated main measurement attribute for {}'.format(name))
        return None
    
    @staticmethod
    def deactivate_lv2_const_optimization(ws, condition):
        print('INFO: Deactivating level 2 constant term optimization for specified pdfs')
        for component in ws.components():
            name = component.GetName()
            if (component.InheritsFrom(ROOT.RooAbsPdf.Class()) and condition(name)):
                component.setAttribute("NOCacheAndTrack")
                print('INFO: Deactivated level 2 constant term optimizatio for {}'.format(name))
    
    @staticmethod
    def get_nuisance_parameter_names(fname:str, ws_name:str, model_config_name:str):
        if not os.path.exists(fname):
            raise FileNotFoundError('workspace file {} does not exist'.format(fname))
        file = ROOT.TFile(fname) 
        if (not file):
            raise RuntimeError("Something went wrong while loading the root file: {}".format(fname))
        ws = file.Get(ws_name)
        if not ws:
            raise RuntimeError("Something went wrong while loading the workspace: {}".format(ws_name))
        model_config = ws.obj(model_config_name)
        if not model_config:
            raise RuntimeError('ERROR: Failed to load model config {}'.format(model_config_name))
        nuisance_parameters = model_config.GetNuisanceParameters()
        if not nuisance_parameters:
            raise RuntimeError('ERROR: Failed to load nuisance parameters')
        return [nuis.GetName() for nuis in nuisance_parameters]
            
    def initialize(self):
        if not os.path.exists(self.fname):
            raise FileNotFoundError('workspace file {} does not exist'.format(self.fname))
        print('Opening file: {}'.format(self.fname))
        file = ROOT.TFile(self.fname) 
        if (not file):
            raise RuntimeError("Something went wrong while loading the root file: {}".format(self.fname))
        print('Loading workspace: {}'.format(self.ws_name))
        ws = file.Get(self.ws_name)
        if not ws:
            raise RuntimeError("Something went wrong while loading the workspace: {}".format(self.ws_name))
        
        # modify interpolation code
        if self.interpolation_code != -1:
            self.modify_interp_codes(ws, self.interpolation_code,
                                     classes=[ROOT.RooStats.HistFactory.FlexibleInterpVar, ROOT.PiecewiseInterpolation])
        
        # activate binned likelihood
        if self.binned_likelihood:
            self.activate_binned_likelihood(ws)
        
        # set main measurement
        if self.tag_as_measurement:
            self.set_measurement(ws, condition=lambda name: name.startswith(self.tag_as_measurement))
                          
        # deactivate level 2 constant term optimization
            self.deactivate_lv2_const_optimization(ws, 
                condition=lambda name: (name.endswith('_mm') and 'mumu_atlas' in name))
                          
        # load model config
        model_config = ws.obj(self.model_config_name)
        if not model_config:
            raise RuntimeError('ERROR: Failed to load model config {}'.format(self.model_config_name))
        print('INFO: Loaded model config "{}" from workspace'.format(self.model_config_name))

        # load pdf
        pdf = model_config.GetPdf()
        if not pdf:
            raise RuntimeError('Failed to load pdf')
        print('INFO: Loaded model pdf "{}" from model config'.format(pdf.GetName()))
             
        # load dataset
        data = ws.data(self.data_name)
        if not data:
            raise RuntimeError('ERROR: Failed to load dataset')
        print('INFO: Loaded dataset "{}" from workspace'.format(data.GetName()))
                
        # load nuisance parameters
        nuisance_parameters = model_config.GetNuisanceParameters()
        if not nuisance_parameters:
            raise RuntimeError('ERROR: Failed to load nuisance parameters')
        print('INFO: Loaded nuisance parameters from model config')
                
        # Load global observables
        global_observables = model_config.GetGlobalObservables()
        if not global_observables:
            raise RuntimeError('ERROR: Failed to load global observables')          
        print('INFO: Loaded global observables from model config')                  
    
        # Load POIs
        pois = model_config.GetParametersOfInterest()
        if not pois:
            raise RuntimeError('ERROR: Failed to load parameters of interest')
        print('INFO: Loaded parameters of interest from model config')
                                  
        # Load observables
        observables = model_config.GetObservables()
        if not observables:
            raise RuntimeError('ERROR: Failed to load observables')     
        print('INFO: Loaded observables from model config')
        
        self._file                = file
        self._workspace           = ws
        self._model_config        = model_config
        self._pdf                 = pdf
        self._data                = data
        self._nuisance_parameters = nuisance_parameters
        self._global_observables  = global_observables
        self._pois                = pois
        self._observables         = observables
                          
        # Load snapshots
        if self.snapshot_names:
            for snapshot_name in self.snapshot_names:
                self._workspace.loadSnapshot(snapshot_name)
                print('INFO: Loaded snapshot "{}"'.format(snapshot_name))
        
        return None
                
    @staticmethod
    def _fix_parameters(source:"ROOT.RooArgSet", param_expr=None, param_str='parameter'):
        '''
            source: parameters instance
            param_expr: 
        '''            
        param_dict = ExtendedModel.parse_param_expr(param_expr)
        return ExtendedModel._set_parameters(source, param_dict, mode='fix', param_str=param_str)           
    
    @staticmethod
    def _profile_parameters(source:"ROOT.RooArgSet", param_expr=None, param_str='parameter'):
        '''
            source: parameters instance
            param_expr: 
        '''                          
        param_dict = ExtendedModel.parse_param_expr(param_expr)
        return ExtendedModel._set_parameters(source, param_dict, mode='free', param_str=param_str)   
    
    def fix_parameters(self, param_expr=None):
        return self._fix_parameters(self.workspace.allVars(), param_expr=param_expr,
                                    param_str='parameter')
    
    def profile_parameters(self, param_expr=None):
        profiled_parameters = self._profile_parameters(self.workspace.allVars(), param_expr=param_expr,
                                                       param_str='parameter') 
        if not profiled_parameters:
            print('Info: No parameters are profiled.')
        return profiled_parameters 
    
    def fix_nuisance_parameters(self, param_expr=None):
        return self._fix_parameters(self.nuisance_parameters, param_expr=param_expr,
                                    param_str='nuisance parameter')
                          
    def fix_parameters_of_interest(self, param_expr=None):
        return self._fix_parameters(self.pois, param_expr=param_expr, param_str='parameter of interest')

    def profile_parameters_of_interest(self, param_expr=None):
        return self._profile_parameters(self.pois, param_expr=param_expr, param_str='parameter of interest')
    
    @staticmethod
    def _set_parameters(source:"ROOT.RooArgSet", param_dict, mode=None, param_str='parameter'):
        set_parameters = []
        available_parameters = [param.GetName() for param in source]
        for name in param_dict:
            if name in available_parameters:
                ExtendedModel._set_parameter(source[name], param_dict[name], mode=mode, param_str=param_str)
                set_parameters.append(source[name])
            else:
                print('WARNING: Parameter "{}" does not exist. No modification will be made.'.format(name))
        return set_parameters
    
    @staticmethod
    def _set_parameter(param, value, mode=None, param_str='parameter'):
        name = param.GetName()
        old_value = param.getVal()
        new_value = old_value
        if isinstance(value, (float, int)):
            new_value = value
        elif isinstance(value, (list, tuple)):
            if len(value) == 3:
                new_value = value[0]
                v_min, v_max = value[1], value[2]
            elif len(value) == 2:
                v_min, v_max = value[0], value[1]
            else:
                raise ValueError('invalid expression for profiling parameter: {}'.format(value))
            # set range
            if (v_min is not None) and (v_max is not None):
                if (new_value < v_min) or (new_value > v_max):
                    new_value = (v_min + v_max)/2
                param.setRange(v_min, v_max)
                print('INFO: Set {} "{}" range to ({},{})'.format(param_str, name, v_min, v_max))
            elif (v_min is not None):
                if (new_value < v_min):
                    new_value = v_min
                # lower bound is zero, if original value is negative, will flip to positive value
                if (v_min == 0) and (old_value < 0):
                    new_value = abs(old_value)
                param.setMin(v_min)
                print('INFO: Set {} "{}" min value to ({},{})'.format(param_str, name, v_min))
            elif (v_max is not None):
                if (new_value > v_max):
                    new_value = v_max
                # upper bound is zero, if original value is positive, will flip to negative value
                if (v_max == 0) and (old_value > 0):
                    new_value = -abs(old_value)                    
                param.setMax(v_max)
                print('INFO: Set {} "{}" max value to ({},{})'.format(param_str, name, v_max))
        if new_value != old_value:
            param.setVal(new_Val)              
            print('INFO: Set {} "{}" value to {}'.format(param_str, name, new_value))
        if mode=='fix':
            param.setConstant(1)
            print('INFO: Fixed {} "{}" at value {}'.format(param_str, name, param.getVal()))
        elif mode=='free':
            param.setConstant(0)
            print('INFO: "{}" = [{}, {}]'.format(name, param.getMin(), param.getMax()))
        return None

    @staticmethod
    def set_parameter_defaults(source:"ROOT.RooArgSet", value=None, error=None, constant=None,
                               remove_range=None, target:List[str]=None):

        for param in source:
            if (not target) or (param.GetName() in target):
                if remove_range:
                    param.removeRange()            
                if value is not None:
                    param.setVal(value)
                if error is not None:
                    param.setError(error)
                if constant is not None:
                    param.setConstant(constant)
        return None
    
    @staticmethod
    def parse_param_expr(param_expr):
        param_dict = {}
        # if parameter expression is not empty string or None
        if param_expr: 
            if isinstance(param_expr, str):
                param_dict = ExtendedModel.parse_param_str(param_expr)
            elif isinstance(param_expr, dict):
                param_dict = param_dict
            else:
                raise ValueError('invalid format for parameter expression: {}'.format(param_expr))
        elif param_expr is None:
        # if param_expr is None, all parameters will be parsed as None by default
            param_dict = {param.GetName():None for param in source}
        return param_dict
                          
    @staticmethod
    def parse_param_str(param_str):
        '''
        Example: "param_1,param_2=0.5,param_3=-1,param_4=1,param_5=0:100,param_6=:100,param_7=0:"
        '''
        param_str = param_str.replace(' ', '')
        param_list = param_str.split(',')
        param_dict = {}
        for param_expr in param_list:
            expr = param_expr.split('=')
            # case only parameter name is given
            if len(expr) == 1:
                param_dict[expr[0]] = None
            # case both parameter name and value is given
            elif len(expr) == 2:
                param_name = expr[0]
                param_value = expr[1]
                # range like expression
                if ':' in param_value:
                    param_range = param_value.split(':')
                    if len(param_range) != 2:
                        raise ValueError('invalid parameter range: {}'.format(param_value))
                    param_min = float(param_range[0]) if param_range[0].isnumeric() else None
                    param_max = float(param_range[1]) if param_range[1].isnumeric() else None
                    param_dict[param_name] = [param_min, param_max]
                elif is_integer(param_value):
                    param_dict[param_name] = int(param_value)
                else:
                    param_dict[param_name] = float(param_value)
            else:
                raise ValueError('invalid parameter expression: {}'.format(param))
        return param_dict
    
    @staticmethod
    def find_unique_prod_components(root_pdf, components, recursion_count=0):
        if (recursion_count > 50):
            raise RuntimeError('find_unique_prod_components detected infinite loop')
        pdf_list = root_pdf.pdfList()
        if pdf_list.getSize() == 1:
            components.add(pdf_list)
            #print('ProdPdf {} is fundamental'.format(pdf_list.at(0).GetName()))
        else:
            for pdf in pdf_list:
                if pdf.ClassName() != 'RooProdPdf':
                    #print('Pdf {} is no RooProdPdf. Adding it.')
                    components.add(pdf)
                    continue
                find_unique_prod_components(pdf, components, recursion_count+1)
    
    def get_all_constraints(self):
        all_constraints = ROOT.RooArgSet()
        cache_name = "CACHE_CONSTR_OF_PDF_{}_FOR_OBS_{}".format(self.pdf.GetName(), 
                     ROOT.RooNameSet(self.data.get()).content())                 
        constr = self.workspace.set(cache_name)
        if constr:
            # retrieve constrains from cache     
            all_constraints.add(constr)
        else:
            # load information needed to determine attributes from ModelConfig 
            obs = deepcopy(self.observables)
            nuis = deepcopy(self.nuisance_parameters)
            all_constraints = self.pdf.getAllConstraints(obs, nuis, ROOT.kFALSE)
            
        # take care of the case where we have a product of constraint terms
        temp_all_constraints = ROOT.RooArgSet(all_constraints.GetName())
        for constraint in all_constraints:
            if constraint.IsA() == ROOT.RooProdPdf.Class():
                buffer = ROOT.RooArgSet()
                ExtendedModel.find_unique_prod_components(constraint, buffer)
                temp_all_constraints.add(buffer)
            else:
                temp_all_constraints.add(constraint)
        return temp_all_constraints
    
    def inspect_constrained_nuisance_parameter(self, nuis, constraints):
        nuis_name = nuis.GetName()
        print('INFO: On nuisance parameter {}'.format(nuis_name))
        nuip_nom = 0.0
        prefit_variation = 1.0
        found_constraint = ROOT.kFALSE
        found_gaussian_constraint = ROOT.kFALSE
        constraint_type = None
        for constraint in constraints:
            constr_name = constraint.GetName()
            if constraint.dependsOn(nuis):
                found_constraint = ROOT.kTRUE
                constraint_type = 'unknown'
                # Loop over global observables to match nuisance parameter and
                # global observable in case of a constrained nuisance parameter
                found_global_observable = ROOT.kFALSE
                for glob_obs in self.global_observables:
                    if constraint.dependsOn(glob_obs):
                        found_global_observable = ROOT.kTRUE
                        # find constraint width in case of a Gaussian
                        if constraint.IsA() == ROOT.RooGaussian.Class():
                            found_gaussian_constraint = ROOT.kTRUE
                            constraint_type = 'gaus'
                            old_sigma_value = 1.0
                            found_sigma = ROOT.kFALSE
                            for server in constraint.servers():
                                if (server != glob_obs) and (server != nuis):
                                    old_sigma_value = server.getVal()
                                    found_sigma = ROOT.kTRUE
                            if math.isclose(old_sigma_value, 1.0, abs_tol=0.001):
                                old_sigma_value = 1.0
                            if not found_sigma:
                                print('INFO: Sigma for pdf {} not found. Uisng 1.0.'.format(constr_name))
                            else:
                                print('INFO: Uisng {} for sigma of pdf {}'.format(old_sigma_value, constr_name))

                            prefit_variation = old_sigma_value
                        elif constraint.IsA() == ROOT.RooPoisson.Class():
                            constraint_type = 'pois'
                            tau = glob_obs.getVal()
                            print('INFO: Found tau {} of pdf'.format(constr_name))
                            prefit_variation = 1. / math.sqrt(tau)
                            print('INFO: Prefit variation is {}'.format(prefit_variation))
                            nuip_nom = 1.0
                            print("INFO: Assume that {} is nominal value of the nuisance parameter".format(nuip_nom))
        return prefit_variation, constraint_type, nuip_nom
        
    def set_initial_errors(self, source:Optional["ROOT.RooArgSet"]=None):
        if not source:
            source = self.nuisance_parameters
    
        all_constraints = self.get_all_constraints()
        for nuis in source:
            nuis_name = nuis.GetName()
            prefit_variation, constraint_type, _ = self.inspect_constrained_nuisance_parameter(nuis, all_constraints)
            if constraint_type=='gaus':
                print('INFO: Changing error of {} from {} to {}'.format(nuis_name, nuis.getError(), prefit_variation))
                nuis.setError(prefit_variation)
                nuis.removeRange()    
        return None
                
                
                
                
                
                
                
                
                
                
                
                
                
                
                
                
                


from typing import Optional, Dict
import ROOT
import numpy as np

     
def set_attrib(obj, **kwargs):
    for key, value in kwargs.items():
        target = obj
        if '.' in key:
            tokens = key.split('.')
            if len(tokens) != 2:
                raise ValueError('maximum of 1 subfield is allowed but {} is given'.format(len(tokens)-1))
            field, key = tokens[0], tokens[1]
            method_name = 'Get' + field
            if hasattr(obj, 'Get' + field):
                target = getattr(obj, method_name)()
            else:
                raise ValueError('{} object does not contain the method {}'.format(type(target), method_name)) 
        method_name = 'Set' + key
        if hasattr(target, 'Set' + key):
            method_name = 'Set' + key
        elif hasattr(target, key):
            method_name = key
        else:
            raise ValueError('{} object does not contain the method {}'.format(type(target), method_name))         
        if value is None:
            getattr(target, method_name)()
        elif isinstance(value, (list, tuple)):
            getattr(target, method_name)(*value)
        else:
            getattr(target, method_name)(value)
    return obj

class StandardGraphMaker(object):
    @staticmethod
    def GraphAsymErrors(n:int, x:np.ndarray, y:np.ndarray, xerr_lo:Optional[np.ndarray]=None,
                        xerr_hi:Optional[np.ndarray]=None, yerr_lo:Optional[np.ndarray]=None,
                        yerr_hi:Optional[np.ndarray]=None, title:str="", xtitle:str="", ytitle:str="",
                        options:Dict={}):
        xerr_hi = xerr_hi if xerr_hi is not None else np.zeros(len(x))
        xerr_lo = xerr_lo if xerr_lo is not None else np.zeros(len(x))
        yerr_hi = yerr_hi if yerr_hi is not None else np.zeros(len(x))
        yerr_lo = yerr_lo if yerr_lo is not None else np.zeros(len(x))
        graph = ROOT.TGraphAsymmErrors(n, x, y, xerr_lo, xerr_hi, yerr_lo, yerr_hi)
        graph.SetTitle(title)
        graph.GetXaxis().SetTitle(xtitle)
        graph.GetYaxis().SetTitle(ytitle)
            
        return set_attrib(graph, **options)
    
    @staticmethod
    def TH2F(name:str, title:str, n_x:int, n_y:int, x_lo:Optional[float]=None,
             x_hi:Optional[float]=None, y_lo:Optional[float]=None,
             y_hi:Optional[float]=None, x_bins: Optional[np.ndarray]=None,
             y_bins: Optional[np.ndarray]=None, options:Dict={}):
        if (x_lo is None) and (x_hi is None) and (y_lo is None) and (y_hi is None):
            hist = ROOT.TH2F(name, title, n_x, x_bins, n_y, y_bins)
        elif (x_lo is None) and (x_hi is None) and (y_bins is None):
            hist = ROOT.TH2F(name, title, n_x, x_bins, n_y, y_lo, y_hi)
        elif (y_lo is None) and (y_hi is None) and (x_bins is None):
            hist = ROOT.TH2F(name, title, n_x, x_lo, x_hi, n_y, y_bins)            
        elif (y_bins is None) and (x_bins is None):
            hist = ROOT.TH2F(name, title, n_x, x_lo, x_hi, n_y, y_lo, y_hi)
        else:
            raise ValueError('cannot convert arguments for TH2F')
        return set_attrib(hist, **options)
    
    @staticmethod
    def TGaxis(x_min:float, y_min:float, x_max:float, y_max:float, w_min:float,
               w_max:float, n_div:int=510, chopt:str="", grid_length:float=0.0,
               options:Dict={}):
        axis = ROOT.TGaxis(x_min, y_min, x_max, y_max, w_min, w_max, n_div, chopt, grid_length)
        return set_attrib(axis, **options)
    
    @staticmethod                 
    def TLine(x1:Optional[float]=None, y1:Optional[float]=None, 
              x2:Optional[float]=None, y2:Optional[float]=None, options:Dict={}):
        if (x1 is None) and (y1 is None) and (x2 is None) and (y2 is None):
            line = ROOT.TLine()
        else:
            line = ROOT.TLine(x1, y1, x2, y2)
        return set_attrib(line, **options)
    
    @staticmethod
    def TLegend(x1:Optional[float]=None, y1:Optional[float]=None, 
                x2:Optional[float]=None, y2:Optional[float]=None,
                w:Optional[float]=None, h:Optional[float]=None,
                header:str="", option:str="", options:Dict={}):
        if (x1 is None) and (y1 is None) and (x2 is None) and (y2 is None):
            legend = ROOT.TLegend(w, h, header, option)
        else:
            legend = ROOT.TLegend(x1, y1, x2, y2, header, option)
        return set_attrib(legend, **options)
    
    @staticmethod
    def TLatex(options:Dict={}):
        latex = ROOT.TLatex()
        return set_attrib(latex, **options)


import re
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
    
VERSIONFILE="quickstats/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))    

setuptools.setup(
    name="quickstats", # Replace with your own username
    version=verstr,
    author="Alkaid Cheng",
    author_email="chi.lung.cheng@cern.ch",
    description="A tool for quick statistical analysis for HEP experiments",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    package_data={'quickstats':['macros/*.so','macros/*.pcm']},
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
          'numpy'
      ],
    scripts=['bin/quickstats'],
    python_requires='>=3.7',
)



from quickstats._version import __version__

import os
import sys
import pathlib

import ROOT
ROOT.PyConfig.IgnoreCommandLineOptions = True
macro_path = pathlib.Path(__file__).parent.absolute()
ROOT.gSystem.Load(os.path.join(macro_path, 'macros', 'RooTwoSidedCBShape_cxx'))
from ROOT import RooTwoSidedCBShape


import os
import time
import json
import fnmatch

import uproot
import numpy as np

from quickstats.components import CLIParser, ExtendedModel, ExtendedMinimizer
from quickstats.utils import root_utils, common_utils

import ROOT

class NuisanceParameterPull(object):
    _CLI_USAGE_          = 'quickstats run_pulls [-h|--help] [<args>]'
    _CLI_DESCRIPTION_    = 'Tool for computing the impact of a given NP to a set of POIs'    
    _CLI_ARG_OPTIONS_    = \
        {
            'input_file': {
                'abbr'        : 'i',
                'description' : 'Path to the input workspace file',
                'required'    : True,
                'type'        : str
            },
            'workspace':{
                'abbr'        : 'w',
                'description' : 'Name of workspace',
                'required'    : False,
                'type'        : str,
                'default'     : 'combWS'
            },
            'model_config':{
                'abbr'        : 'm',
                'description' : 'Name of model config',
                'required'    : False,
                'type'        : str,
                'default'     : 'ModelConfig'
            },  
            'data':{
                'abbr'        : 'd',
                'description' : 'Name of dataset',
                'required'    : False,
                'type'        : str,
                'default'     : 'combData'
            },
            'parameter'  : {
                'abbr'        : 'p',
                'description' : 'Nuisance parameter(s) to run pulls on.' +\
                                'Multiple parameters are separated by commas.' +\
                                'Wildcards are accepted', 
                'required'    : False,
                'type'        : str,
                'default'     : ''
            },   
            'poi': {
                'abbr'        : 'x',
                'description' : 'POIs to measure',
                'required'    : False,
                'type'        : str,
                'default'     : ""
            },
            'profile': {
                'abbr'        : 'r',
                'description' : 'Parameters to profile', 
                'required'    : False,
                'type'        : str,
                'default'     : ""
            },
            'fix': {
                'abbr'        : 'f',
                'description' : 'Parameters to fix', 
                'required'    : False,
                'type'        : str,
                'default'     : ""
            },
            'snapshot': {
                'abbr'        : 's',
                'description' : 'Name of initial snapshot', 
                'required'    : False,
                'type'        : str,
                'default'     : "nominalNuis"
            },
            'outdir': {
                'abbr'        : 'o',
                'description' : 'Output directory', 
                'required'    : False,
                'type'        : str,
                'default'     : "output"
            },
            'minimizer_type'  : {
                'abbr'        : 't',
                'description' : 'Minimizer type', 
                'required'    : False,
                'type'        : str,
                'default'     : "Minuit2"
            },        
            'minimizer_algo'  : {
                'abbr'        : 'a',
                'description' : 'Minimizer algorithm', 
                'required'    : False,
                'type'        : str,
                'default'     : "Migrad"
            },          
            'num_cpu'  : {
                'abbr'        : 'c',
                'description' : 'Number of CPUs to use', 
                'required'    : False,
                'type'        : int,
                'default'     : 1
            },   
            'binned'  : {
                'abbr'        : 'b',
                'description' : 'Binned likelihood', 
                'required'    : False,
                'type'        : int,
                'default'     : 1
            },    
            'precision'  : {
                'abbr'        : 'q',
                'description' : 'Precision for scan', 
                'required'    : False,
                'type'        : float,
                'default'     : 0.001
            },          
            'eps'  : {
                'abbr'        : 'e',
                'description' : 'Convergence criterium', 
                'required'    : False,
                'type'        : float,
                'default'     : 1.0
            },              
            'log_level'  : {
                'abbr'        : 'l',
                'description' : 'Log level', 
                'required'    : False,
                'type'        : str,
                'default'     : "INFO"
            },      
            'eigen'  : {
                'abbr'        : None,
                'description' : 'Compute eigenvalues and vectors', 
                'required'    : False,
                'type'        : int,
                'default'     : 0
            },   
            'strategy'  : {
                'abbr'        : None,
                'description' : 'default strategy', 
                'required'    : False,
                'type'        : int,
                'default'     : 0
            },
            'fix_cache'  : {
                'abbr'        : None,
                'description' : 'Fix StarMomentMorph cache', 
                'required'    : False,
                'type'        : int,
                'default'     : 1
            },    
            'fix_multi'  : {
                'abbr'        : None,
                'description' : 'Fix MultiPdf level 2', 
                'required'    : False,
                'type'        : int,
                'default'     : 1
            },            
            'offset'  : {
                'abbr'        : None,
                'description' : 'Offset likelihood', 
                'required'    : False,
                'type'        : int,
                'default'     : 1
            },  
            'optimize'  : {
                'abbr'        : None,
                'description' : 'Optimize constant terms', 
                'required'    : False,
                'type'        : int,
                'default'     : 2
            },       
            'max_calls'  : {
                'abbr'        : None,
                'description' : 'Maximum number of function calls', 
                'required'    : False,
                'type'        : int,
                'default'     : -1
            }, 
            'max_iters'  : {
                'abbr'        : None,
                'description' : 'Maximum number of Minuit iterations', 
                'required'    : False,
                'type'        : int,
                'default'     : -1
            },       
            'parallel'  : {
                'abbr'        : None,
                'description' : 'Parallelize job across different nuisance parameters using N workers.'+\
                                'Use -1 for N_CPU workers.', 
                'required'    : False,
                'type'        : int,
                'default'     : 0
            },                 
        } 
    
    @property
    def model(self):
        return self._model
    
    @property
    def workspace(self):
        return self._workspace
    @property
    def model_config(self):
        return self._model_config
    @property
    def pdf(self):
        return self._pdf
    @property
    def data(self):
        return self._data
    @property
    def nuisance_parameters(self):
        return self._nuisance_parameters
    @property
    def global_observables(self):
        return self._global_observables
    @property
    def pois(self):
        return self._pois
    @property
    def observables(self):
        return self._observables  
    
    def __init__(self):
        self._model               = None
        self._workspace           = None
        self._model_config        = None
        self._pdf                 = None
        self._data                = None
        self._nuisance_parameters = None
        self._global_observables  = None
        self._pois                = None
        self._observables         = None
    
    def get_parser(self, **kwargs):
        parser = CLIParser(description=self._CLI_DESCRIPTION_,
                           usage=self._CLI_USAGE_)
        parser.load_argument_options(self._CLI_ARG_OPTIONS_)
        return parser
    
    def run_parser(self, args=None):
        parser = self.get_parser()
        kwargs = vars(parser.parse_args(args))
        self.run_pulls(**kwargs)
    
    @staticmethod
    def evaluate_impact(model:ExtendedModel, minimizer:ExtendedMinimizer,
                        nuis, nominal_value, pois, minimizer_options, snapshot=None):
        poi_values = []
        start_time = time.time()
        if snapshot:
            model.workspace.loadSnapshot(snapshot)
        nuis.setVal(nominal_value)
        nuis.setConstant(1)   
        minimizer.minimize(**minimizer_options)
        for poi in pois:
            poi_values.append(poi.getVal())
        elapsed_time = time.time() - start_time
        return poi_values, elapsed_time
    
    @staticmethod
    def _run_pulls(input_file:str, workspace:str, model_config:str, data:str,
                   snapshot:str, nuis_name:str, poi_name:str, 
                   fix_param:str='', profile_param:str='', binned:bool=True,
                   minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', 
                   precision:float=0.001, eps:float=1.0, strategy:int=0, 
                   num_cpu:int=1, offset:int=1, optimize:int=2, eigen:bool=False,
                   fix_cache:bool=True, fix_multi:bool=True, max_calls:int=-1, 
                   max_iters:int=-1, outdir:str='output', **kwargs):
        start_time = time.time()
        snapshot_names = snapshot.split(',')
        # load model
        model = ExtendedModel(model_name="model", fname=input_file, ws_name=workspace,
                              model_config_name=model_config, data_name=data, 
                              snapshot_names=snapshot_names, binned_likelihood=binned,
                              tag_as_measurement="pdf_", fix_cache=fix_cache, fix_multi=fix_multi)
        if fix_param:
            model.fix_parameters(fix_param)
            
        # by default fix all POIs before floating
        model.set_parameter_defaults(model.pois, error=0.15, constant=1, remove_range=True)
        for param in model.pois:
            extra_str = 'Fixing' if param.isConstant() else 'Set'
            print('INFO: {} POI {} at value {}'.format(extra_str, param.GetName(), param.getVal()))
        
        # collect pois
        rank_pois = model.profile_parameters(poi_name)
        model.set_parameter_defaults(rank_pois, error=0.15)     
        
        # profile pois
        if profile_param:
            print('INFO: Profiling POIs')
            profile_pois = model.profile_parameters(profile)
        
        buffer_time = time.time()
    
        nuip = model.workspace.var(nuis_name)
        if not nuip:
            raise ValueError('Nuisance parameter "{}" does not exist'.format(parameter))
        nuip_name = nuip.GetName()
        nuip.setConstant(0)
        print('INFO: Computing error for parameter "{}" at {}'.format(nuip.GetName(), nuip.getVal()))
        
        print("INFO: Making ExtendedMinimizer for unconditional fit")
        minimizer = ExtendedMinimizer("minimizer", model.pdf, model.data)
        print("INFO: Starting minimization")
        nll_commands = [ROOT.RooFit.NumCPU(num_cpu, 3), 
                        ROOT.RooFit.Constrain(model.nuisance_parameters),
                        ROOT.RooFit.GlobalObservables(model.global_observables), 
                        ROOT.RooFit.Offset(offset)]

        minimize_options = {
            'minimizer_type'   : minimizer_type,
            'minimizer_algo'   : minimizer_algo,
            'default_strategy' : strategy,
            'opt_const'        : optimize,
            'precision'        : precision,
            'eps'              : eps,
            'max_calls'        : max_calls,
            'max_iters'        : max_iters,
        }

        minimizer.minimize(nll_commands=nll_commands,
                           scan=1, scan_set=ROOT.RooArgSet(nuip),
                           **minimize_options)
        unconditional_time = time.time() - buffer_time
        print("INFO: Fitting time: {:.3f} s".format(unconditional_time))
        pois_hat = []
        for rank_poi in rank_pois:
            name = rank_poi.GetName()
            value = rank_poi.getVal()
            pois_hat.append(value)
            print('{} {}'.format(name, value))
        
        model.workspace.saveSnapshot('tmp_snapshot', model.pdf.getParameters(model.data))
        print('INFO: Made unconditional snapshot with name tmp_snapshot')
        
        # find prefit variation
        buffer_time = time.time()
        
        nuip_hat = nuip.getVal()
        nuip_errup = nuip.getErrorHi()
        nuip_errdown = nuip.getErrorLo()

        all_constraints = model.get_all_constraints()
        prefit_variation, constraint_type, nuip_nom = model.inspect_constrained_nuisance_parameter(nuip, all_constraints)
        if not constraint_type:
            print('INFO: Not a constrained parameter. No prefit impact can be determined. Use postfit impact instead.')
        prefit_uncertainty_time = time.time() - buffer_time
        print('INFO: Time to find prefit variation: {:.3f} s'.format(prefit_uncertainty_time))
        
        if rank_pois:
            new_minimizer_options = {
                'nll_commands': nll_commands,
                'reuse_nll'   : 1,
                **minimize_options
            }
            # fix theta at the MLE value +/- postfit uncertainty and minimize again to estimate the change in the POI
            print('INFO: Evaluating effect of moving {} up by {} + {}'.format(nuip_name, nuip_hat, nuip_errup))
            pois_up, postfit_up_impact_time = NuisanceParameterPull.evaluate_impact(model, minimizer,
                                                nuip, nuip_hat + abs(nuip_errup), rank_pois,
                                                new_minimizer_options,  'tmp_snapshot')
            print('INFO: Time to find postfit up impact: {:.3f} s'.format(postfit_up_impact_time))
            
            print('INFO: Evaluating effect of moving {} down by {} - {}'.format(nuip_name, nuip_hat, nuip_errup))
            pois_down, postfit_down_impact_time = NuisanceParameterPull.evaluate_impact(model, minimizer,
                                                    nuip, nuip_hat - abs(nuip_errdown), rank_pois,
                                                    new_minimizer_options,  'tmp_snapshot')
            print('INFO: Time to find postfit down impact: {:.3f} s'.format(postfit_down_impact_time))
            
            # fix theta at the MLE value +/- prefit uncertainty and minimize again to estimate the change in the POI
            
            if constraint_type:
                print('Evaluating effect of moving {} up by {} + {}'.format(nuip_name, nuip_hat, prefit_variation))
                pois_nom_up, prefit_up_impact_time = NuisanceParameterPull.evaluate_impact(model, minimizer,
                                                        nuip, nuip_hat + prefit_variation, rank_pois,
                                                        new_minimizer_options,  'tmp_snapshot')
                print('INFO: Time to find prefit up impact: {:.3f} s'.format(prefit_up_impact_time))      
                
                print('Evaluating effect of moving {} down by {} - {}'.format(nuip_name, nuip_hat, prefit_variation))
                pois_nom_down, prefit_down_impact_time = NuisanceParameterPull.evaluate_impact(model, minimizer,
                                                            nuip, nuip_hat - prefit_variation, rank_pois,
                                                            new_minimizer_options,  'tmp_snapshot')
                print('INFO: Time to find prefit down impact: {:.3f} s'.format(prefit_up_impact_time))
            else:
                print('WARNING: Prefit impact not estimated, instead postfit impact is cloned')
                pois_nom_up = [i for i in pois_up]
                pois_nom_down = [i for i in pois_down]
        else:
            pois_up, pois_down, pois_nom_up, pois_nom_down = [], [], [], []
        
        end_time = time.time()
        print('\nINFO: Time to perform all fits: {:.3f} s'.format(end_time-start_time))
        print('INFO: Unconditional minimum of NP {}: {} + {} - {}'.format(nuis_name, nuip_hat, 
              abs(nuip_errup), abs(nuip_errdown)))
        print('INFO: Prefit uncertainy of NP {}: {} +/- {}'.format(nuis_name, nuip_hat, prefit_variation))
        for i, rank_poi in enumerate(rank_pois):
            print('INFO: Unconditional minimum of POI {}: {}'.format(rank_poi.GetName(), pois_hat[i]))
            print('INFO: POI when varying NP up by 1 sigma postfit (prefit): {} ({})'.format(pois_up[i], pois_nom_up[i]))
            print('INFO: POI when varying NP down by 1 sigma postfit (prefit): {} ({})'.format(pois_down[i], pois_nom_down[i]))

        # create output directory if not exists
        pulls_out_dir = os.path.join(outdir, 'pulls')
        if not os.path.exists(outdir):
            os.makedirs(outdir)
            
        # store result in root file
        outname_root = os.path.join(outdir, nuis_name + '.root')
        
        result = {}
        result['nuis'] = {  'nuisance'   : nuis_name,
                            'nuis_nom'   : nuip_nom,
                            'nuis_hat'   : nuip_hat,
                            'nuis_hi'    : nuip_errup,
                            'nuis_lo'    : nuip_errdown,
                            'nuis_prefit': prefit_variation}
        result['pois'] = {}
        for i, rank_poi in enumerate(rank_pois):
            name = rank_poi.GetName()
            result['pois'][name] = { 'hat'     : pois_hat[i],
                                     'up'      : pois_up[i],
                                     'down'    : pois_down[i],
                                     'up_nom'  : pois_nom_up[i],
                                     'down_nom': pois_nom_down[i]}
            
        result_for_root = {}
        result_for_root.update(result['nuis'])
        for k,v in result['pois'].items():
            buffer = {'{}_{}'.format(k, kk): vv for kk,vv in v.items()}
            result_for_root.update(buffer)
        r_file = ROOT.TFile(outname_root, "RECREATE")
        r_tree = ROOT.TTree("result", "result")
        root_utils.fill_branch(r_tree, result_for_root)
        r_file.Write()
        r_file.Close()
        print('INFO: Saved output to {}'.format(outname_root))
        outname_json = os.path.join(outdir, nuis_name + '.json')
        json.dump(result, open(outname_json, 'w'), indent=2)
        

        
    def run_pulls(self, input_file='workspace.root', workspace='combWS', model_config='ModelConfig',
                 data='combData', poi='', snapshot='nominalNuis', outdir='output', profile="",
                 fix='', minimizer_type='Minuit2', minimizer_algo='Migrad', num_cpu=1, binned=1,
                 precision=0.001, eps=1.0, log_level='INFO', eigen=0, strategy=0, fix_cache=1, fix_multi=1,
                 offset=1, optimize=2, parameter="", max_calls=-1, max_iters=-1, parallel=0, **kwargs):
        
        start_time = time.time()
        
        # configure default minimizer options
        ExtendedMinimizer._configure_default_minimizer_options(minimizer_type, minimizer_algo,
            strategy, debug_mode=(log_level=="DEBUG"))
        
        
        nuis_list = ExtendedModel.get_nuisance_parameter_names(input_file, workspace, model_config)
        nuis_patterns = parameter.split(',')
        if parameter:
            nuis_names = []
            for nuis_name in nuis_list:
                # filter out nuisance parameters
                if any(fnmatch.fnmatch(nuis_name, nuis_pattern) for nuis_pattern in nuis_patterns):
                    nuis_names.append(nuis_name)
        else:
            nuis_names = nuis_list
                
        if parallel:
            from itertools import repeat
            arguments = (repeat(input_file), repeat(workspace), repeat(model_config),
                         repeat(data), repeat(snapshot), nuis_names, repeat(poi),
                         repeat(fix), repeat(profile), repeat(binned),
                         repeat(minimizer_type), repeat(minimizer_algo), repeat(precision),
                         repeat(eps), repeat(strategy), repeat(num_cpu), repeat(offset),
                         repeat(optimize), repeat(eigen), repeat(fix_cache), repeat(fix_multi),
                         repeat(max_calls), repeat(max_iters), repeat(outdir))
            max_workers = parallel if parallel != -1 else common_utils.get_cpu_count()
            common_utils.parallel_run(self._run_pulls, *arguments, max_workers=max_workers)
        else:
            for nuis_name in nuis_names:
                self._run_pulls(input_file, workspace, model_config, data, snapshot,
                                nuis_name, poi, fix, profile, binned, minimizer_type, 
                                minimizer_algo, precision, eps, strategy, num_cpu, offset, 
                                optimize, eigen, fix_cache, fix_multi, max_calls, max_iters, outdir)
        end_time = time.time()
        print('INFO: All jobs have finished. Total Time taken: {:.3f} s'.format(end_time-start_time))
        
    @staticmethod
    def parse_root_result(fname):
        with uproot.open(fname) as file:
            result = root_utils.uproot_to_dict(file)
        return result


import numpy as np

def is_integer(s:str):
    if not s:
        return False
    if len(s) == 1:
        return s.isdigit()
    if s[0] in ('-', '+'):
        return s[1:].isdigit()
    return s.isdigit()

def array_swap(arr1:np.ndarray, arr2:np.ndarray, indices):
    arr1[indices], arr2[indices] = arr2[indices], arr1[indices]
    
def reorder_arrays(*arrays, descending:bool=True):
    if descending:
        if not (arrays[0].dtype.type in [np.string_, np.str_]):
            indices = np.argsort(-arrays[0])
        else:
            indices = np.argsort(arrays[0])[::-1]
    else:
        indices = np.argsort(arrays[0])
    for arr in arrays:
        arr[:] = arr[indices]    
        
        
def reverse_arrays(*arrays):
    for arr in arrays:
        arr[:] = arr[::-1] 